/*
 *  Canon Inkjet Printer Driver for Linux
 *  Copyright CANON INC. 2022
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
*/

#include <stdio.h>
#include <string.h>
#include "com_log.h"
#include "com_util.h"

// Functions
/* =============================================================================
  [Function Name]
    utilCreateOption
  [Outline]
    Create option
  [Parameters]
    const char        *key        [in]            key to add
    const char        *value      [in]            value to add
  [Return]
    Created option
============================================================================= */
option_list *utilCreateOption(const char *key, const char *value)
{
  int key_len = strlen(key) + 1;
  int value_len = strlen(value) + 1;
  option_list *root = (option_list *)malloc(sizeof(option_list));

  memset(root, 0, sizeof(option_list));

  root->key = malloc(key_len);
  memset(root->key, 0, key_len);
  strcpy(root->key, key);

  root->value = malloc(value_len);
  memset(root->value, 0, value_len);
  strcpy(root->value, value);

  return root;
}

/* =============================================================================
  [Function Name]
    utilFindOption
  [Outline]
    Find option from list
  [Parameters]
    option_list       *root       [in]            option list
    const char        *key        [in]            key to find
  [Return]
    Return the found option, NULL for Not found
============================================================================= */
option_list *utilFindOption(option_list *root, const char *key)
{
  option_list *curs;

  for (curs = root; curs != NULL; curs = curs->next)
  {
    if (!strcmp(curs->key, key))
    {
      return curs;
    }
  }

  return NULL;
}

/* =============================================================================
  [Function Name]
    utilAddOption
  [Outline]
    Add option to list
  [Parameters]
    option_list       *root       [out]           option list
    const char        *key        [in]            key to add
    const char        *value      [in]            value to add
============================================================================= */
void utilAddOption(option_list *root, const char *key, const char *value)
{
  option_list *curs = NULL;
  option_list *last = NULL;

  for (curs = root; curs != NULL; last = curs, curs = curs->next)
  {
    if (!strcmp(curs->key, key))
    {
      break;
    }
  }

  if (curs == NULL)
  {
    if (last != NULL)
    {
      last->next = utilCreateOption(key, value);
    }
  }
  else
  {
    if (strcmp(curs->value, value))
    {
      utilUpdateOption(curs, value);
    }
  }
}

/* =============================================================================
  [Function Name]
    utilUpdateOption
  [Outline]
    Update option
  [Parameters]
    option_list       *option     [in]            option list
    const char        *value      [in]            value to update
============================================================================= */
void utilUpdateOption(option_list *option, const char *value)
{
  int value_len = strlen(value) + 1;

  free(option->value);
  option->value = malloc(value_len);
  memset(option->value, 0, value_len);
  strcpy(option->value, value);
}

/* =============================================================================
  [Function Name]
    utilFreeOptionList
  [Outline]
    Free the option list
  [Parameters]
    option_list         **root      [out]           option list
============================================================================= */
void utilFreeOptionList(option_list *root)
{
  option_list *next;
  option_list *pl = root;
  for (; pl != NULL; pl = next)
  {
    next = pl->next;

    free(pl->key);
    free(pl->value);
    free(pl);
  }
}

/* =============================================================================
  [Function Name]
    utilFindOptionValue
  [Outline]
    Find option value from list
  [Parameters]
    option_list       *root       [in]            option list
    const char        *key        [in]            key to find
  [Return]
    Return the found option value, NULL for Not found
============================================================================= */
const char *utilFindOptionValue(option_list *root, const char *key)
{
  option_list *curs = utilFindOption(root, key);
  const char *result = NULL;
 
  if (curs != NULL)
  {
    result = curs->value;
  }

  return result;
}
