/*
 *  CUPS add-on module for Canon Inkjet Printer.
 *  Copyright CANON INC. 2022
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
*/

#include "image.h"
#include "com_log.h"

/* =============================================================================
  [Function Name]
    createImageScaler
  [Outline]
    Create image scaler
  [Parameters]
    cups_page_header2_t     *cupsHeader     [in]          CUPS header for current page
    int                     out_w           [in]          Output image width
    int                     out_h           [in]          Output image height
  [Return]
    image_scaler pointer
============================================================================= */
image_scaler *createImageScaler(cups_page_header2_t *cupsHeader, int out_w, int out_h)
{
  double coef_w, coef_h;
  int in_w, in_h, dst_w, dst_h, ofs_w, ofs_h;
  image_scaler *imageScaler = NULL;

  in_w = cupsHeader->cupsWidth;
  in_h = cupsHeader->cupsHeight;
  if (in_w == 0 || in_h == 0 || out_w == 0 || out_h == 0)
  {
    return imageScaler;
  }

  coef_w = (double)out_w / (double)in_w;
  coef_h = (double)out_h / (double)in_h;

  if (coef_w < coef_h)
  {
    dst_w = out_w;
    dst_h = (int)((double)in_h * coef_w + 0.5 < 1 ? 1 : (double)in_h * coef_w + 0.5);
    ofs_w = 0;
    ofs_h = (out_h - dst_h) / 2;
  }
  else
  {
    dst_w = (int)((double)in_w * coef_h + 0.5 < 1 ? 1 : (double)in_w * coef_h + 0.5);
    dst_h = out_h;
    ofs_w = (out_w - dst_w) / 2;
    ofs_h = 0;
  }

  imageScaler = malloc(sizeof(image_scaler));

  imageScaler->srcWidth = in_w;
  imageScaler->srcHeight = in_h;
  imageScaler->srcData = malloc(in_w * cupsHeader->cupsNumColors);
  memset(imageScaler->srcData, 0, in_w * cupsHeader->cupsNumColors);
  imageScaler->numColors = cupsHeader->cupsNumColors;
  imageScaler->dstWidth = dst_w;
  imageScaler->dstHeight = dst_h;
  imageScaler->marginWidth = ofs_w;
  imageScaler->marginHeight = ofs_h;
  imageScaler->quotientW = dst_w > 1 ? (in_w - 1) / (dst_w - 1) : 0;
  imageScaler->restW = dst_w > 1 ? (in_w - 1) % (dst_w - 1) : 0;
  imageScaler->quotientH = dst_h > 1 ? (in_h - 1) / (dst_h - 1) : 0;
  imageScaler->restH = dst_h > 1 ? (in_h - 1) % (dst_h - 1) : 0;
  imageScaler->currPosH = 0;
  imageScaler->prevPosH = -1;
  imageScaler->totalRestH = 0;
  imageScaler->totalReadLines = 0;

  DEBUG_PRINT_D("[canonijlfp] createImageScaler(): imageScaler is created\n");
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->srcWidth = %d\n", imageScaler->srcWidth);
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->srcHeight = %d\n", imageScaler->srcHeight);
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->numColors = %d\n", imageScaler->numColors);
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->dstWidth = %d\n", imageScaler->dstWidth);
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->dstHeight = %d\n", imageScaler->dstHeight);
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->marginWidth = %d\n", imageScaler->marginWidth);
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->marginHeight = %d\n", imageScaler->marginHeight);
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->quotientW = %d\n", imageScaler->quotientW);
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->restW = %d\n", imageScaler->restW);
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->quotientH = %d\n", imageScaler->quotientH);
  DEBUG_PRINT2_D("[canonijlfp] imageScaler->restH = %d\n", imageScaler->restH);
  return imageScaler;
}

/* =============================================================================
  [Function Name]
    setImageScalerIO
  [Outline]
    Set image scaler IO function to read bytes
  [Parameters]
    image_scaler            *imageScaler    [in]        Scale information
    cups_raster_iocb_t      iocb            [in]        Read callback
    void                    *ctx            [in]        Context pointer for callback
============================================================================= */
void setImageScalerIO(image_scaler *imageScaler, cups_raster_iocb_t iocb, void *ctx)
{
  imageScaler->iocb.iocb = iocb;
  imageScaler->iocb.ctx = ctx;
}

/* =============================================================================
  [Function Name]
    scaleImageData
  [Outline]
    scale RasterData
  [Parameters]
    image_scaler          *imageScaler    [in]        Scale information
    int                   lines           [in]        Destination lines
    uint8_t               *data           [out]       Scaled raster data
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int scaleImageData(image_scaler *imageScaler, int lines, uint8_t *data)
{
  int h, w, i;
  int totalRestH, currPosH, prevPosH;
  uint8_t *dstData = NULL, *srcData = NULL;
  int srcBufferSize;

  DEBUG_PRINT_D("[canonijlfp] Enter scaleImageData()\n");

  if (imageScaler == NULL || imageScaler->iocb.iocb == NULL || data == NULL)
  {
    DEBUG_PRINT_D("[canonijlfp] Leave scaleImageData()\n");
    return 1;
  }

  srcBufferSize = imageScaler->numColors * imageScaler->srcWidth;
  currPosH = imageScaler->currPosH;
  prevPosH = imageScaler->prevPosH;
  totalRestH = imageScaler->totalRestH;
  dstData = data;
  for (h = 0; h < lines; h++)
  {
    int totalRestW = 0;
    if ((imageScaler->dstHeight > 1) && (totalRestH * 2 >= (imageScaler->dstHeight - 1)))
    {
      totalRestH -= (imageScaler->dstHeight - 1);
      currPosH++;
    }

    for (i = 0; i < currPosH - prevPosH; i++)
    {
      memset(imageScaler->srcData, WHITE, srcBufferSize);
      imageScaler->iocb.iocb(imageScaler->iocb.ctx, imageScaler->srcData, srcBufferSize);
    }

    srcData = imageScaler->srcData;
    for (w = 0; w < imageScaler->dstWidth; w++)
    {
      if ((imageScaler->dstWidth > 1) && (totalRestW * 2 >= (imageScaler->dstWidth - 1)))
      {
        totalRestW -= (imageScaler->dstWidth - 1);
        srcData += imageScaler->numColors;
      }

      memcpy(dstData, srcData, imageScaler->numColors);

      totalRestW += imageScaler->restW;
      dstData += imageScaler->numColors;
      srcData += imageScaler->quotientW * imageScaler->numColors;
    }

    imageScaler->totalReadLines += currPosH - prevPosH;
    prevPosH = currPosH;
    currPosH += imageScaler->quotientH;
    totalRestH += imageScaler->restH;
  }

  imageScaler->currPosH = currPosH;
  imageScaler->prevPosH = prevPosH;
  imageScaler->totalRestH = totalRestH;

  DEBUG_PRINT_D("[canonijlfp] Succeed to invoke scaleImageData()\n");
  DEBUG_PRINT_D("[canonijlfp] Leave scaleImageData()\n");
  return 0;
}

/* =============================================================================
  [Function Name]
    destroyImageScaler
  [Outline]
    Destroy image scaler
  [Parameters]
    image_scaler          *imageScaler       [in]        Scale information
============================================================================= */
void destroyImageScaler(image_scaler *imageScaler)
{
  if (imageScaler)
  {
    free(imageScaler->srcData);
    free(imageScaler);
    DEBUG_PRINT_D("[canonijlfp] destroyImageScaler(): imageScaler is deleted\n");
  }
}