/*
 *  CUPS add-on module for Canon Inkjet Printer.
 *  Copyright CANON INC. 2022
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
*/

#pragma GCC diagnostic ignored "-Wdeprecated-declarations"

#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <getopt.h>
#include <dlfcn.h>
#include <cups/cups.h>
#include <cups/raster.h>

#include "com_def.h"
#include "com_log.h"
#include "com_util.h"
#include "libcnivec.h"
#include "libcnsgraster.h"
#include "image.h"

#define POINT_DPI           72
#define INCH_TO_MM_100      2540
#define MAX_BAND_LINE       256
#define MAX_DATA_SIZE       (1024 * 1024)
#define IS_SUPPORT          1
#define MAX_OPTION_LENGTH   256

// Variables definiation
int g_cancel_job = 0;

void *g_lib_ivec = NULL;
fnIvecSendStartJob ivecSendStartJob = NULL;
fnIvecSendJobConfiguration ivecSendJobConfiguration = NULL;
fnIvecSendConfiguration ivecSendConfiguration = NULL;
fnIvecSendSendData ivecSendSendData = NULL;
fnIvecSendEndJob ivecSendEndJob = NULL;

void *g_lib_sgraster = NULL;
fnSgrCreateEncoder sgrCreateEncoder = NULL;
fnSgrDestroyEncoder sgrDestroyEncoder = NULL;
fnSgrStartImage sgrStartImage = NULL;
fnSgrEncodeImage sgrEncodeImage = NULL;
fnSgrFinishImage sgrFinishImage = NULL;

// Functions definiation
void sigtermHandler(int sigcode);
option_list *generateOptionList(int argc, char *argv[]);
int loadLibraryIvec();
int loadLibrarySgraster();
int sendOnePage(unsigned int printArea[], cups_page_header2_t *cupsHeader, cups_raster_t *ras, option_list *options, FILE *fp);
int sendAllPages(option_list *options, FILE *fp);

struct option supportOptList[] = {
    {SETTING_JOBNAME, required_argument, NULL, IS_SUPPORT},
    {SETTING_USERNAME, required_argument, NULL, IS_SUPPORT},
    {SETTING_COMPUTERNAME, required_argument, NULL, IS_SUPPORT},
    {SETTING_PAPERSIZE, required_argument, NULL, IS_SUPPORT},
    {SETTING_PAPERSIZE_CUSTOM_WIDTH, required_argument, NULL, IS_SUPPORT},
    {SETTING_PAPERSIZE_CUSTOM_HEIGHT, required_argument, NULL, IS_SUPPORT},
    {SETTING_PRINTAREA_LEFT, required_argument, NULL, IS_SUPPORT},
    {SETTING_PRINTAREA_TOP, required_argument, NULL, IS_SUPPORT},
    {SETTING_PRINTAREA_WIDTH, required_argument, NULL, IS_SUPPORT},
    {SETTING_PRINTAREA_HEIGHT, required_argument, NULL, IS_SUPPORT},
    {SETTING_PAPERTYPE, required_argument, NULL, IS_SUPPORT},
    {SETTING_INPUTBIN, required_argument, NULL, IS_SUPPORT},
    {SETTING_PRINTTYPE, required_argument, NULL, IS_SUPPORT},
    {SETTING_COLORMODE, required_argument, NULL, IS_SUPPORT},
    {SETTING_COLORMODE_INTENT, required_argument, NULL, IS_SUPPORT},
    {SETTING_PRINTQUALITY, required_argument, NULL, IS_SUPPORT},
    {SETTING_PAPERSAVE, required_argument, NULL, IS_SUPPORT},
    {0, 0, 0, 0},
};

// Functions
/* =============================================================================
  [Function Name]
    main
  [Outline]
    Main function
  [Parameters]
    int           argc        [in]        arguments count
    char          *argv[]     [in]        arguments
  [Return]
    PROC_SUCCEEDED: Succeed
    PROC_FAILED: Failed
============================================================================= */
int main(int argc, char *argv[])
{
  struct sigaction sigact;
  FILE *fp = stdout;
  int ret = PROC_FAILED;
  option_list *options = NULL;

  DEBUG_PRINT_PROGRAM_START();

  // Ignore SIGPIPE
  memset(&sigact, 0, sizeof(sigact));
  sigact.sa_handler = SIG_IGN;
  if (sigaction(SIGPIPE, &sigact, NULL))
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to register SIGPIPE process method\n");
    goto EXIT;
  }

  // Set SIGTERM
  memset(&sigact, 0, sizeof(sigact));
  sigact.sa_handler = sigtermHandler;
  if (sigaction(SIGTERM, &sigact, NULL))
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to register SIGTERM process method\n");
    goto EXIT;
  }

  if (loadLibraryIvec())
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to invoke loadLibraryIvec()\n");
    goto EXIT;
  }

  if (loadLibrarySgraster())
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to invoke loadLibrarySgraster()\n");
    goto EXIT;
  }

  options = generateOptionList(argc, argv);
  if (ivecSendStartJob(options, fp) == IVEC_FAIL)
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to invoke ivecSendStartJob()\n");
    goto EXIT;
  }

  if (ivecSendJobConfiguration(options, fp) == IVEC_FAIL)
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to invoke ivecSendJobConfiguration()\n");
    goto EXIT;
  }

  if (sendAllPages(options, fp))
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to invoke sendAllPages()\n");
    goto EXIT;
  }

  if (ivecSendEndJob(options, fp) == IVEC_FAIL)
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to invoke ivecSendEndJob()\n");
    goto EXIT;
  }

  ret = PROC_SUCCEEDED;

EXIT:
  if (options != NULL)
  {
    utilFreeOptionList(options);
  }

  if (g_lib_ivec)
  {
    dlclose(g_lib_ivec);
  }

  if (g_lib_sgraster)
  {
    dlclose(g_lib_sgraster);
  }

  DEBUG_PRINT_PROGRAM_END();
  return ret;
}

/* =============================================================================
  [Function Name]
    sendOnePage
  [Outline]
    Send one page data
  [Parameters]
    unsigned int          printArea[]     [in]        print area
    cups_page_header2_t   *cupsHeader     [in]        CUPS header for current page
    cups_raster_t         *ras            [in]        CUPS raster data for current page
    option_list           *options        [in]        option list
    FILE                  *fp             [in]        output port
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int sendOnePage(unsigned int printArea[], cups_page_header2_t *cupsHeader, cups_raster_t *ras, option_list *options, FILE *fp)
{
  int ret = 1;
  page_header pageHeader;
  uint8_t *data = NULL;
  sgrencoder *enc = NULL;
  encoder_settings ens;
  int i, bandCount;
  image_scaler *imageScaler;
  int outputHeight;

  DEBUG_PRINT_D("[canonijlfp] Enter sendOnePage()\n");

  // Compute output page size and margin
  imageScaler = createImageScaler(cupsHeader, printArea[2], printArea[3]);
  if (imageScaler == NULL)
  {
    DEBUG_PRINT_D("[canonijlfp] Failed to invoke createImageScaler()\n");
    goto EXIT;
  }
  setImageScalerIO(imageScaler, (cups_raster_iocb_t)cupsRasterReadPixels, ras);

  outputHeight = printArea[4] > imageScaler->dstHeight ? imageScaler->dstHeight : printArea[4];
  bandCount = (outputHeight + MAX_BAND_LINE - 1) / MAX_BAND_LINE;

  pageHeader.width = imageScaler->dstWidth;
  pageHeader.bytesPerLine = imageScaler->numColors * imageScaler->dstWidth;
  pageHeader.margins[0] = imageScaler->marginWidth;
  pageHeader.margins[1] = imageScaler->marginHeight;
  pageHeader.resolution = cupsHeader->HWResolution[0];

  ens.maxBandLines = MAX_BAND_LINE;
  ens.maxDataSize = MAX_DATA_SIZE;
  ens.fnSendData = ivecSendSendData;
  ens.fp = fp;

  DEBUG_PRINT2_D("[canonijlfp] maxBandLines = %u\n", ens.maxBandLines);
  DEBUG_PRINT2_D("[canonijlfp] maxDataSize = %u\n", ens.maxDataSize);
  DEBUG_PRINT2_D("[canonijlfp] bandCount = %d\n", bandCount);

  DEBUG_PRINT2_D("[canonijlfp] cupsHeader.cupsWidth = %d\n", cupsHeader->cupsWidth);
  DEBUG_PRINT2_D("[canonijlfp] cupsHeader.cupsHeight = %d\n", cupsHeader->cupsHeight);
  DEBUG_PRINT2_D("[canonijlfp] cupsHeader.cupsImagingBBox[0] = %f\n", cupsHeader->cupsImagingBBox[0]);
  DEBUG_PRINT2_D("[canonijlfp] cupsHeader.cupsImagingBBox[1] = %f\n", cupsHeader->cupsImagingBBox[1]);
  DEBUG_PRINT2_D("[canonijlfp] cupsHeader.cupsImagingBBox[2] = %f\n", cupsHeader->cupsImagingBBox[2]);
  DEBUG_PRINT2_D("[canonijlfp] cupsHeader.cupsImagingBBox[3] = %f\n", cupsHeader->cupsImagingBBox[3]);
  DEBUG_PRINT2_D("[canonijlfp] cupsHeader.HWResolution[0] = %d\n", cupsHeader->HWResolution[0]);
  DEBUG_PRINT2_D("[canonijlfp] cupsHeader.HWResolution[1] = %d\n", cupsHeader->HWResolution[1]);

  DEBUG_PRINT2_D("[canonijlfp] pageHeader.width = %u\n", pageHeader.width);
  DEBUG_PRINT2_D("[canonijlfp] pageHeader.margins[0] = %u\n", pageHeader.margins[0]);
  DEBUG_PRINT2_D("[canonijlfp] pageHeader.margins[1] = %u\n", pageHeader.margins[1]);
  DEBUG_PRINT2_D("[canonijlfp] pageHeader.resolution = %u\n", pageHeader.resolution);

  if (!g_cancel_job && ivecSendConfiguration(options, fp) == IVEC_FAIL)
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to invoke ivecSendConfiguration()\n");
    goto EXIT;
  }

  enc = sgrCreateEncoder(&pageHeader, &ens, options);
  if (enc == NULL)
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to invoke sgrCreateEncoder()\n");
    goto EXIT;
  }

  if (!g_cancel_job && sgrStartImage(enc) == IVEC_FAIL)
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to invoke sgrStartImage()\n");
    goto EXIT;
  }

  data = malloc(pageHeader.bytesPerLine * MAX_BAND_LINE);
  if (data == NULL)
  {
    DEBUG_PRINT_E("[canonijlfp] Not enough memory\n");
    goto EXIT;
  }

  for (i = 0; !g_cancel_job && i < bandCount; i++)
  {
    int lines = 0;

    DEBUG_PRINT2_D("[canonijlfp] Process band %d\n", i);

    lines = outputHeight - i * MAX_BAND_LINE;
    lines = lines > MAX_BAND_LINE ? MAX_BAND_LINE : lines;
    memset(data, 0, pageHeader.bytesPerLine * MAX_BAND_LINE);
    if (scaleImageData(imageScaler, lines, data))
    {
      DEBUG_PRINT_E("[canonijlfp] Failed to invoke scaleImageData()\n");
      goto EXIT;
    }

    if (sgrEncodeImage(enc, data, lines) == IVEC_FAIL)
    {
      DEBUG_PRINT_E("[canonijlfp] Failed to invoke sgrEncodeImage()\n");
      goto EXIT;
    }
  }

  for (i = 0; !g_cancel_job && i < (int)cupsHeader->cupsHeight - imageScaler->totalReadLines; i++)
  {
    cupsRasterReadPixels(ras, data, cupsHeader->cupsBytesPerLine);
  }

  if (!g_cancel_job && sgrFinishImage(enc) == IVEC_FAIL)
  {
    DEBUG_PRINT_E("[canonijlfp] Failed to invoke sgrFinishImage()\n");
    goto EXIT;
  }

  ret = 0;
  DEBUG_PRINT_D("[canonijlfp] Send page successfully\n");

EXIT:
  if (data)
  {
    free(data);
  }

  if (imageScaler)
  {
    destroyImageScaler(imageScaler);
  }

  if (enc)
  {
    sgrDestroyEncoder(enc);
  }

  DEBUG_PRINT_D("[canonijlfp] Leave sendOnePage()\n");
  return ret;
}

/* =============================================================================
  [Function Name]
    sendAllPages
  [Outline]
    Send all pages data
  [Parameters]
    option_list   *options        [in]        option list
    FILE          *fp             [in]        output port
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int sendAllPages(option_list *options, FILE *fp)
{
  int ret = 0;
  int fd = 0;
  cups_raster_t *ras = NULL;
  cups_page_header2_t header;
  const char *valueAreaLeft = utilFindOptionValue(options, SETTING_PRINTAREA_LEFT);
  const char *valueAreaTop = utilFindOptionValue(options, SETTING_PRINTAREA_TOP);
  const char *valueAreaWidth = utilFindOptionValue(options, SETTING_PRINTAREA_WIDTH);
  const char *valueAreaHeight = utilFindOptionValue(options, SETTING_PRINTAREA_HEIGHT);
  const char *valueCustomWidth = utilFindOptionValue(options, SETTING_PAPERSIZE_CUSTOM_WIDTH);
  const char *valueCustomHeight = utilFindOptionValue(options, SETTING_PAPERSIZE_CUSTOM_HEIGHT);
  int imageAreaLeft, imageAreaTop, imageAreaWidth, imageAreaHeight, pageAreaWidth, pageAreaHeight;

  DEBUG_PRINT_D("[canonijlfp] Enter sendAllPages()\n");

  if (!valueAreaLeft || !valueAreaTop || !valueAreaWidth || !valueAreaHeight) {
    DEBUG_PRINT_E("[canonijlfp] Failed to get printer area from parameter!\n");
    DEBUG_PRINT_D("[canonijlfp] Leave sendAllPages()\n");
    return 1;
  }

  // image area
  imageAreaLeft = atoi(valueAreaLeft);
  imageAreaTop = atoi(valueAreaTop);
  imageAreaWidth = atoi(valueAreaWidth);
  imageAreaHeight = atoi(valueAreaHeight);

  // page area
  pageAreaWidth = valueCustomWidth ? atoi(valueCustomWidth) - imageAreaLeft * 2 : imageAreaWidth;
  pageAreaHeight = valueCustomHeight ? atoi(valueCustomHeight) - imageAreaTop * 2 : imageAreaHeight;

  ras = cupsRasterOpen(fd, CUPS_RASTER_READ);
  while (!g_cancel_job && !ret && cupsRasterReadHeader2(ras, &header))
  {
    // adjust the height of the print area in IVEC by following items
    // 1. media position is PPD, which indicate to different input_bin
    // 2. fix CINC-issue006, which cause the image is cut when custom page size is less than the minimize size
    int adjustedAreaHeight = imageAreaTop - header.MediaPosition + ((pageAreaHeight > imageAreaHeight) ? pageAreaHeight : imageAreaHeight);
    char mediaTypeString[MAX_OPTION_LENGTH] = {0};

    unsigned int printArea[] = {
      imageAreaLeft * header.HWResolution[0] / INCH_TO_MM_100,
      imageAreaTop * header.HWResolution[1] / INCH_TO_MM_100,
      imageAreaWidth * header.HWResolution[0] / INCH_TO_MM_100,
      imageAreaHeight * header.HWResolution[1] / INCH_TO_MM_100,
      adjustedAreaHeight * header.HWResolution[1] / INCH_TO_MM_100
    };

    // update print area width
    if (imageAreaWidth != pageAreaWidth)
    {
      char areaWidthString[MAX_OPTION_LENGTH] = {0};
      sprintf(areaWidthString, "%d", pageAreaWidth);
      utilAddOption(options, SETTING_PRINTAREA_WIDTH, areaWidthString);
      DEBUG_PRINT3_D("[canonijlfp] Update width of print area: %d => %d\n", imageAreaWidth, pageAreaWidth);
    }

    // update print area height
    if (imageAreaHeight != adjustedAreaHeight)
    {
      char areaHeightString[MAX_OPTION_LENGTH] = {0};
      sprintf(areaHeightString, "%d", adjustedAreaHeight);
      utilAddOption(options, SETTING_PRINTAREA_HEIGHT, areaHeightString);
      DEBUG_PRINT3_D("[canonijlfp] Update height of print area: %d => %d\n", imageAreaHeight, adjustedAreaHeight);
    }

    // add papertype
    sprintf(mediaTypeString, "%d", header.cupsMediaType);
    utilAddOption(options, SETTING_PAPERTYPE, mediaTypeString);

    DEBUG_PRINT_D("[canonijlfp] Processing Page...\n");
    ret = sendOnePage(printArea, &header, ras, options, fp);
  }
  
  cupsRasterClose(ras);
  close(fd);

  DEBUG_PRINT_D("[canonijlfp] Leave sendAllPages()\n");
  return ret;
}

/* =============================================================================
  [Function Name]
    loadLibraryIvec
  [Outline]
    Load IVEC library
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int loadLibraryIvec()
{
  g_lib_ivec = dlopen(LIBNAME_IVEC, RTLD_LAZY);
  if (!g_lib_ivec) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to open libaray: %s\n", LIBNAME_IVEC);
    return 1;
  }

  ivecSendStartJob = dlsym(g_lib_ivec, EXPORT_IVECSENDSTARTJOB);
  if (dlerror() != NULL) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to get libaray function: %s\n", EXPORT_IVECSENDSTARTJOB);
    return 1;
  }

  ivecSendJobConfiguration = dlsym(g_lib_ivec, EXPORT_IVECSENDJOBCONFIGURATION);
  if (dlerror() != NULL) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to get libaray function: %s\n", EXPORT_IVECSENDJOBCONFIGURATION);
    return 1;
  }

  ivecSendConfiguration = dlsym(g_lib_ivec, EXPORT_IVECSENDCONFIGURATION);
  if (dlerror() != NULL) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to get libaray function: %s\n", EXPORT_IVECSENDCONFIGURATION);
    return 1;
  }

  ivecSendSendData = dlsym(g_lib_ivec, EXPORT_IVECSENDSENDDATA);
  if (dlerror() != NULL) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to get libaray function: %s\n", EXPORT_IVECSENDSENDDATA);
    return 1;
  }

  ivecSendEndJob = dlsym(g_lib_ivec, EXPORT_IVECSENDENDJOB);
  if (dlerror() != NULL) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to get libaray function: %s\n", EXPORT_IVECSENDENDJOB);
    return 1;
  }

  return 0;
}

/* =============================================================================
  [Function Name]
    loadLibrarySgraster
  [Outline]
    Load SGRaster library
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int loadLibrarySgraster()
{
  g_lib_sgraster = dlopen(LIBNAME_SGRASTER, RTLD_LAZY);
  if (!g_lib_sgraster) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to open libaray: %s\n", LIBNAME_SGRASTER);
    return 1;
	}

  sgrCreateEncoder = dlsym(g_lib_sgraster, EXPORT_SGRCREATEENCODER);
  if (dlerror() != NULL) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to get libaray function: %s\n", EXPORT_SGRCREATEENCODER);
    return 1;
  }

  sgrDestroyEncoder = dlsym(g_lib_sgraster, EXPORT_SGRDESTROYENCODER);
  if (dlerror() != NULL) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to get libaray function: %s\n", EXPORT_SGRDESTROYENCODER);
    return 1;
  }

  sgrStartImage = dlsym(g_lib_sgraster, EXPORT_SGRSTARTIMAGE);
  if (dlerror() != NULL) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to get libaray function: %s\n", EXPORT_SGRSTARTIMAGE);
    return 1;
  }

  sgrEncodeImage = dlsym(g_lib_sgraster, EXPORT_SGRENCODEIMAGE);
  if (dlerror() != NULL) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to get libaray function: %s\n", EXPORT_SGRENCODEIMAGE);
    return 1;
  }

  sgrFinishImage = dlsym(g_lib_sgraster, EXPORT_SGRFINISHIMAGE);
  if (dlerror() != NULL) {
    DEBUG_PRINT2_E("[canonijlfp] Failed to get libaray function: %s\n", EXPORT_SGRFINISHIMAGE);
    return 1;
  }

  return 0;
}

/* =============================================================================
  [Function Name]
    generateOptionList
  [Outline]
    Generate option list
  [Parameters]
    int           argc        [in]        arguments count
    char          *argv[]     [in]        arguments
  [Return]
    option_list pointer
============================================================================= */
option_list *generateOptionList(int argc, char *argv[])
{
  option_list *options = utilCreateOption(SETTING_SERVICETYPE, SETTING_SERVICETYPE_PRINT);
  int optIndex, value;
  while ((value = getopt_long(argc, argv, "", supportOptList, &optIndex)) != -1)
  {
    if (value == IS_SUPPORT)
    {
      DEBUG_PRINT3_D("[canonijlfp] OPTION(%s):VALUE(%s)\n", supportOptList[optIndex].name, optarg);
      utilAddOption(options, supportOptList[optIndex].name, optarg);
    }
  }

  return options;
}

/* =============================================================================
  [Function Name]
    sigtermHandler
  [Outline]
    Process SIGTERM action
  [Parameters]
    int           sigcode       [in]         signal code
  [Return]
    Sub process ID
============================================================================= */
void sigtermHandler(int sigcode)
{
  g_cancel_job = 1;

  DEBUG_PRINT_D("[canonijlfp] sigtermHandler(): The process is terminated\n");
}