/*
 *  CUPS add-on module for Canon Inkjet Printer.
 *  Copyright CANON INC. 2022
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
*/

#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <unistd.h>

#include "com_def.h"
#include "com_log.h"
#include "libcnivec.h"

#define IS_SUPPORT          1
#define DATA_BUF_SIZE       1024
#define CMD_TEST_PRINT      0
#define CMD_CLEANING        1
#define CMD_AUTO_ALIGNMENT  2

#define UTIL_CMD_SUFFIX           "#CUPS-COMMAND"
#define UTIL_CMD_TEST_PRINT       "PrintSelfTestPage"
#define UTIL_CMD_CLEANING         "Clean all"
#define UTIL_CMD_AUTO_ALIGNMENT   "com.canon.autoalignment"

// Variables definiation
int g_cancel_job = 0;

void *g_lib_ivec = NULL;
fnIvecSendStartJob ivecSendStartJob = NULL;
fnIvecSendJobConfiguration ivecSendJobConfiguration = NULL;
fnIvecSendCleaning ivecSendCleaning = NULL;
fnIvecSendTestPrint ivecSendTestPrint = NULL;
fnIvecSendEndJob ivecSendEndJob = NULL;

// Functions definiation
void sigtermHandler(int sigcode);
option_list *generateOptionList();
int loadLibraryIvec();
int getMaintenanceType(int ifd, int *type);
int sendCupsCommand(option_list *options, FILE *fp, int dType);
int read_line(int fd, char *p_buf, int bytes);

// Functions
/* =============================================================================
  [Function Name]
    main
  [Outline]
    Main function
  [Parameters]
    int           argc        [in]        arguments count
    char          *argv[]     [in]        arguments
  [Return]
    PROC_SUCCEEDED: Succeed
    PROC_FAILED: Failed
============================================================================= */
int main(int argc, char *argv[])
{
  struct sigaction sigact;
  FILE *fp = stdout;
  int ret = PROC_FAILED;
  option_list *options = NULL;
  int ifd = 0;
  int dType = -1;

  DEBUG_PRINT_PROGRAM_START();

  // Ignore SIGPIPE
  memset(&sigact, 0, sizeof(sigact));
  sigact.sa_handler = SIG_IGN;
  if (sigaction(SIGPIPE, &sigact, NULL))
  {
    DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to register SIGPIPE process method\n");
    goto EXIT;
  }

  // Set SIGTERM
  memset(&sigact, 0, sizeof(sigact));
  sigact.sa_handler = sigtermHandler;
  if (sigaction(SIGTERM, &sigact, NULL))
  {
    DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to register SIGTERM process method\n");
    goto EXIT;
  }

  if (argc < 6 || argc > 7)
  {
    DEBUG_PRINT_E("[cmdtocanonijlfp] Illegal parameter number.\n");
    goto EXIT;
  }

  if (argc == 7)
  {
    if ((ifd = open(argv[6], O_RDONLY)) == -1)
    {
      DEBUG_PRINT_E("[cmdtocanonijlfp] Cannot open file.\n");
      goto EXIT;
    }
  }

  // get maintenance command type
  if (getMaintenanceType(ifd, &dType))
  {
    DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to invoke getMaintenanceType()\n");
    goto EXIT;
  }

  if (loadLibraryIvec())
  {
    DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to invoke loadLibraryIvec()\n");
    goto EXIT;
  }

  options = generateOptionList();
  if (ivecSendStartJob(options, fp) == IVEC_FAIL)
  {
    DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to invoke ivecSendStartJob()\n");
    goto EXIT;
  }

  if (ivecSendJobConfiguration(options, fp) == IVEC_FAIL)
  {
    DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to invoke ivecSendJobConfiguration()\n");
    goto EXIT;
  }

  if (sendCupsCommand(options, fp, dType))
  {
    DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to invoke sendCupsCommand()\n");
    goto EXIT;
  }

  if (ivecSendEndJob(options, fp) == IVEC_FAIL)
  {
    DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to invoke ivecSendEndJob()\n");
    goto EXIT;
  }

  ret = PROC_SUCCEEDED;

EXIT:
  if (options != NULL)
  {
    utilFreeOptionList(options);
  }

  if (g_lib_ivec)
  {
    dlclose(g_lib_ivec);
  }

  if (ifd > 0)
  {
    close(ifd);
  }

  DEBUG_PRINT_PROGRAM_END();
  return ret;
}

/* =============================================================================
  [Function Name]
    read_line
  [Outline]
    Read file line by line
  [Parameters]
    int           fd              [in]        file descriptor
    char          *p_buf          [out]       result buffer
    int           bytes           [in]        buffer size
  [Return]
    read bytes number
============================================================================= */
int read_line(int fd, char *p_buf, int bytes)
{
  static char read_buf[DATA_BUF_SIZE];
  static int buf_bytes = 0;
  static int buf_pos = 0;
  int read_bytes = 0;

  memset(p_buf, 0, bytes);
  while (bytes > 0)
  {
    if (buf_bytes == 0 && fd != -1)
    {
      buf_bytes = read(fd, read_buf, DATA_BUF_SIZE);
      if (buf_bytes > 0)
      {
        buf_pos = 0;
      }
    }

    if (buf_bytes > 0)
    {
      *p_buf = read_buf[buf_pos++];
      bytes--;
      buf_bytes--;
      read_bytes++;

      if (*p_buf == '\n')
      {
        break;
      }
      p_buf++;
    }
    else if (buf_bytes < 0)
    {
      if (errno == EINTR)
      {
        continue;
      }
    }
    else
    {
      break;
    }
  }

  return read_bytes;
}

/* =============================================================================
  [Function Name]
    getMaintenanceType
  [Outline]
    Get Maintenance Type
  [Parameters]
    int           ifd             [in]        file descriptor
    int           *type           [out]       maintenance type
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int getMaintenanceType(int ifd, int *type)
{
  char read_buf[DATA_BUF_SIZE];
  int read_bytes;
  int rtn = 1;
  int lines = 0;

  memset(read_buf, 0, sizeof(read_buf));

  for (lines = 0; (read_bytes = read_line(ifd, read_buf, DATA_BUF_SIZE - 1)) > 0; lines++)
  {
    // check cups command file suffix.
    if (lines == 0)
    {
      if (strncmp(read_buf, UTIL_CMD_SUFFIX, read_bytes - 1) != 0)
      {
        DEBUG_PRINT_E("[cmdtocanonijlfp] Command file data are invalid.\n");
        goto error_exit;
      }
      else
      {
        continue;
      }
    }

    // perform only first command.
    if (strncmp(read_buf, UTIL_CMD_TEST_PRINT, strlen(read_buf) - 1) == 0)
    {
      *type = CMD_TEST_PRINT;
      break;
    }
    else if (strncmp(read_buf, UTIL_CMD_CLEANING, strlen(read_buf) - 1) == 0)
    {
      *type = CMD_CLEANING;
      break;
    }
    else if (strncmp(read_buf, UTIL_CMD_AUTO_ALIGNMENT, strlen(read_buf) - 1) == 0)
    {
      *type = CMD_AUTO_ALIGNMENT;
      break;
    }
    else
    {
      DEBUG_PRINT_E("[cmdtocanonijlfp] Cannot detect command.\n");
      goto error_exit;
    }
  }

  rtn = 0;

error_exit:
  return rtn;
}

/* =============================================================================
  [Function Name]
    sendCupsCommand
  [Outline]
    Send CUPS-COMMAND
  [Parameters]
    option_list   *options        [in]        option list
    FILE          *fp             [in]        output port
    int           dType           [in]        maintenance type
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int sendCupsCommand(option_list *options, FILE *fp, int dType)
{
  int ret = 0;

  DEBUG_PRINT_D("[cmdtocanonijlfp] Enter sendCupsCommand()\n");

  if (dType == CMD_TEST_PRINT)
  {
    utilAddOption(options, SETTING_MAINTENANCE_TYPE, SETTING_MAINTENANCE_TYPE_TEST_PRINT);

    if (ivecSendTestPrint(options, fp) == IVEC_FAIL)
    {
      DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to invoke ivecSendTestPrint()\n");
      return 1;
    }
  }
  else if (dType == CMD_AUTO_ALIGNMENT)
  {
    utilAddOption(options, SETTING_MAINTENANCE_TYPE, SETTING_MAINTENANCE_TYPE_AUTO_ALIGNMENT);

    if (ivecSendTestPrint(options, fp) == IVEC_FAIL)
    {
      DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to invoke ivecSendTestPrint()\n");
      return 1;
    }
  }
  else if (dType == CMD_CLEANING)
  {
    if (ivecSendCleaning(options, fp) == IVEC_FAIL)
    {
      DEBUG_PRINT_E("[cmdtocanonijlfp] Failed to invoke ivecSendCleaning()\n");
      return 1;
    }
  }

  DEBUG_PRINT_D("[cmdtocanonijlfp] Leave sendCupsCommand()\n");
  return ret;
}

/* =============================================================================
  [Function Name]
    loadLibraryIvec
  [Outline]
    Load IVEC library
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int loadLibraryIvec()
{
  g_lib_ivec = dlopen(LIBNAME_IVEC, RTLD_LAZY);
  if (!g_lib_ivec)
  {
    DEBUG_PRINT2_E("[cmdtocanonijlfp] Failed to open libaray: %s\n", LIBNAME_IVEC);
    return 1;
  }

  ivecSendStartJob = dlsym(g_lib_ivec, EXPORT_IVECSENDSTARTJOB);
  if (dlerror() != NULL)
  {
    DEBUG_PRINT2_E("[cmdtocanonijlfp] Failed to get libaray function: %s\n", EXPORT_IVECSENDSTARTJOB);
    return 1;
  }

  ivecSendJobConfiguration = dlsym(g_lib_ivec, EXPORT_IVECSENDJOBCONFIGURATION);
  if (dlerror() != NULL)
  {
    DEBUG_PRINT2_E("[cmdtocanonijlfp] Failed to get libaray function: %s\n", EXPORT_IVECSENDJOBCONFIGURATION);
    return 1;
  }

  ivecSendCleaning = dlsym(g_lib_ivec, EXPORT_IVECSENDCLEANING);
  if (dlerror() != NULL)
  {
    DEBUG_PRINT2_E("[cmdtocanonijlfp] Failed to get libaray function: %s\n", EXPORT_IVECSENDCLEANING);
    return 1;
  }

  ivecSendTestPrint = dlsym(g_lib_ivec, EXPORT_IVECSENDTESTPRINT);
  if (dlerror() != NULL)
  {
    DEBUG_PRINT2_E("[cmdtocanonijlfp] Failed to get libaray function: %s\n", EXPORT_IVECSENDTESTPRINT);
    return 1;
  }

  ivecSendEndJob = dlsym(g_lib_ivec, EXPORT_IVECSENDENDJOB);
  if (dlerror() != NULL)
  {
    DEBUG_PRINT2_E("[cmdtocanonijlfp] Failed to get libaray function: %s\n", EXPORT_IVECSENDENDJOB);
    return 1;
  }

  return 0;
}

/* =============================================================================
  [Function Name]
    generateOptionList
  [Outline]
    Generate option list
  [Return]
    option_list pointer
============================================================================= */
option_list *generateOptionList()
{
  option_list *options = utilCreateOption(SETTING_SERVICETYPE, SETTING_SERVICETYPE_MAINTENANCE);

  utilAddOption(options, SETTING_JOBNAME, "");
  utilAddOption(options, SETTING_USERNAME, "");
  utilAddOption(options, SETTING_COMPUTERNAME, "");

  return options;
}

/* =============================================================================
  [Function Name]
    sigtermHandler
  [Outline]
    Process SIGTERM action
  [Parameters]
    int           sigcode       [in]         signal code
  [Return]
    Sub process ID
============================================================================= */
void sigtermHandler(int sigcode)
{
  g_cancel_job = 1;

  DEBUG_PRINT_D("[cmdtocanonijlfp] sigtermHandler(): The process is terminated\n");
}