/*
 *  CUPS add-on module for Canon Inkjet Printer.
 *  Copyright CANON INC. 2022
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
*/

#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include "com_def.h"
#include "com_log.h"
#include "common.h"

#define CANONIJLFP_BIN                "canonijlfp"
#define CMD_HEADER                    "--"
#define PAGESIZE_CUSTOM               "Custom"

#define BUF_SIZE                      256
#define DATA_BUF_SIZE                 (1024 * 256)
#define PAGESIZE_DPI                  72
#define PAGESIZE_INCH_TO_MM_10        254
#define PAGESIZE_RATIO_MM_10          10
#define INTERGER_MAX                  10
#define FLOAT_EPSILON                 0.1f


const char *optionArray[] = {
  SETTING_PAPERSIZE,
  SETTING_INPUTBIN,
  SETTING_PRINTQUALITY,
  SETTING_COLORMODE,
  SETTING_COLORMODE_INTENT,
  SETTING_PRINTTYPE,
  SETTING_PAPERSAVE
};

// Functions
/* =============================================================================
  [Function Name]
    addOption
  [Outline]
    add option
  [Parameters]
    const char *name              [in]        option name
    const char *value             [in]        option value
    int num_options               [in]        option number
    cups_option_t **options       [out]       output option
  [Return]
    option number
============================================================================= */
int addOption(const char *name, const char *value, int num_options, cups_option_t **options)
{
  int ret = num_options;
  char *inputName;
  int strLen = strlen(name) + strlen(CMD_HEADER);

  if((inputName = malloc(strLen + 1)) == NULL)
  {
    return ret;
  }

  memset(inputName, 0, strLen + 1);
  strcpy(inputName, CMD_HEADER);
  strcat(inputName, name);

  ret = cupsAddOption(inputName, value, num_options, options);
  free(inputName);
  return ret;
}

/* =============================================================================
  [Function Name]
    getPrintSettings
  [Outline]
    Get print settings
  [Parameters]
    char          *arg            [in]        arguments from main process
    cups_option_t **options       [out]       options to transfer to sub process
    int           *num_opt        [out]       options count
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int getPrintSettings(char *arg, cups_option_t **options, int *num_opt)
{
  ppd_file_t *ppd;
  char *ppdName = getenv("PPD");
  int numOut = 0;
  cups_option_t *optionIn = NULL;
  cups_option_t *optionOut = NULL;
  int i;

  DEBUG_PRINT_D("[rastertocanonijlfp] Enter getPrintSettings()\n");
  
  //Parse argv and ppd options
  if((ppd = ppdOpenFile( ppdName )) == NULL)
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] ppdOpenFile() error\n");
    DEBUG_PRINT_D("[rastertocanonijlfp] Leave getPrintSettings()\n");
    return 1;
  }
  ppdMarkDefaults(ppd);

  if(arg != NULL)
  {
    int numIn = cupsParseOptions(arg, 0, &optionIn);
    cupsMarkOptions(ppd, numIn, optionIn);
    cupsFreeOptions(numIn, optionIn);
  }
  
  for(i = 0; i < sizeof(optionArray)/sizeof(optionArray[0]); i++)
  {
    ppd_choice_t	*choice = ppdFindMarkedChoice(ppd, optionArray[i]);
    if(choice != NULL)
    {
      numOut = addOption(optionArray[i], choice->choice, numOut, &optionOut);
    }
  }

  if (setPrintAreaOption(ppd, &optionOut, &numOut))
  {
    fprintf(stderr, "ERROR: Failed to print due to the custom paper size is out of support!\n");
    ppdClose(ppd);
    return 1;
  }

  *num_opt = numOut;
  *options = optionOut;

  ppdClose(ppd);

  DEBUG_PRINT2_D("[rastertocanonijlfp] %d options are added\n", *num_opt);
  DEBUG_PRINT_D("[rastertocanonijlfp] Leave getPrintSettings()\n");
  return 0;
}

/* =============================================================================
  [Function Name]
    setPrintAreaOption
  [Outline]
    Set Printarea option
  [Parameters]
    ppd_file_t          *ppd            [in]        ppd file
    cups_option_t       **options       [out]       options to transfer to sub process
    int                 *num_opt        [out]       options count
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int setPrintAreaOption(ppd_file_t *ppd, cups_option_t **options, int *num_opt)
{
  ppd_size_t *size;
  ppd_size_t size_cus_min;
  ppd_size_t size_cus_max;
  char pageSizeStr[BUF_SIZE] = {0};
  // 0:left 1:top 2:width 3:height 4:custom_width 5:custom_height
  float pageSize[6];
  int itemCount = 0;
  char *ivecKeys[] = {SETTING_PRINTAREA_LEFT, SETTING_PRINTAREA_TOP, SETTING_PRINTAREA_WIDTH, SETTING_PRINTAREA_HEIGHT,
                      SETTING_PAPERSIZE_CUSTOM_WIDTH, SETTING_PAPERSIZE_CUSTOM_HEIGHT};
  int i;

  snprintf(pageSizeStr, BUF_SIZE, "%s%s", CMD_HEADER, SETTING_PAPERSIZE);
  size = ppdPageSize(ppd, cupsGetOption(pageSizeStr, *num_opt, *options));

  // if smaller than margin, return failed
  if (size->right < size->left || size->top < size->bottom)
  {
    return 1;
  }

  pageSize[itemCount++] = size->left;
  pageSize[itemCount++] = size->bottom;
  pageSize[itemCount++] = size->right - size->left;
  pageSize[itemCount++] = size->top - size->bottom;

  // check if custom paper
  if (0 == strcmp(size->name, PAGESIZE_CUSTOM))
  {
    float max_width = 0.0f;
    float max_length = 0.0f;

    // get custom paper size
    if (ppdPageSizeLimits(ppd, &size_cus_min, &size_cus_max) == 0)
    {
      return 1;
    }

    // fix accuracy issue when convert unit from MM to PT
    max_width = ceil(size_cus_max.width) + FLOAT_EPSILON;
    max_length = ceil(size_cus_max.length) + FLOAT_EPSILON;

    // if larger than max, return failed
    if (size->width > max_width || size->length > max_length)
    {
      DEBUG_PRINT3_D("[rastertocanonijlfp] paper width: %f  paper length: %f\n", size->width, size->length);
      DEBUG_PRINT3_D("[rastertocanonijlfp] max_width: %f  max_length: %f\n", max_width, max_length);
      DEBUG_PRINT_D("[rastertocanonijlfp] Paper size exceeds the max size\n");
      return 1;
    }

    // if smaller than min, set to min
    pageSize[itemCount++] = size->width < size_cus_min.width ? size_cus_min.width : size->width;
    pageSize[itemCount++] = size->length < size_cus_min.length ? size_cus_min.length : size->length;
  }

  for (i = 0; i < itemCount; i++)
  {
    int sizeNum = 0;
    char szBuffer[INTERGER_MAX + 1] = {0};
    sizeNum = (int)(round(pageSize[i] / PAGESIZE_DPI * PAGESIZE_INCH_TO_MM_10)) * PAGESIZE_RATIO_MM_10;

    snprintf(szBuffer, INTERGER_MAX, "%d", sizeNum);
    *num_opt = addOption(ivecKeys[i], szBuffer, *num_opt, options);
  }

  return 0;
}

/* =============================================================================
  [Function Name]
    getAdditionalPrintSettings
  [Outline]
    Get addtional print settings
  [Parameters]
    char          *userName       [in]        userName from arguments
    char          *jobName        [in]        jobName from arguments
    cups_option_t **options       [out]       options to transfer to sub process
    int           *num_opt        [out]       options count
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int getAdditionalPrintSettings(char *userName, char *jobName, cups_option_t **options, int *num_opt)
{
  int num = *num_opt;
  char outputJobName[SETTING_JOBNAME_MAXLENGTH + 1] = {0};
  char outputUserName[SETTING_USERNAME_MAXLENGTH + 1] = {0};
  char outputComputerName[SETTING_COMPUTERNAME_MAXLENGTH + 1] = {0};

  DEBUG_PRINT_D("[rastertocanonijlfp] Enter getAdditionalPrintSettings()\n");

  //computername
  if(gethostname(outputComputerName, SETTING_COMPUTERNAME_MAXLENGTH) != 0)
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] Failed to get computer name\n");
    DEBUG_PRINT_D("[rastertocanonijlfp] Leave getAdditionalPrintSettings()\n");
    return 1;
  }
  num = addOption(SETTING_COMPUTERNAME, outputComputerName, num, options);

  //jobname
  strncpy(outputJobName, jobName, SETTING_JOBNAME_MAXLENGTH);
  num = addOption(SETTING_JOBNAME, outputJobName, num, options);

  //username
  strncpy(outputUserName, userName, SETTING_USERNAME_MAXLENGTH);
  num = addOption(SETTING_USERNAME, outputUserName, num, options);
  
  *num_opt = num;

  DEBUG_PRINT2_D("[rastertocanonijlfp] %d options are added\n", *num_opt);
  DEBUG_PRINT_D("[rastertocanonijlfp] Leave getAdditionalPrintSettings()\n");
  return 0;
}

/* =============================================================================
  [Function Name]
    makeExecCommand
  [Outline]
    Make command line for sub process
  [Parameters]
    char          *arg[]          [out]       command line data
    cups_option_t *options        [in]        options to set to commmand line
    int           num_opt         [in]        options count
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int makeExecCommand(char *arg[], cups_option_t *options, int num_opt)
{
  int i, count = 0;

  arg[count++] = CANONIJLFP_BIN;
  for(i = 0; i < num_opt; i++)
  {
    arg[count++] = options[i].name;
    arg[count++] = options[i].value;
  }

  return 0;
}

/* =============================================================================
  [Function Name]
    execFilter
  [Outline]
    Create PIPE and start sub process
  [Parameters]
    char          *arg[]          [in]        command line data
    int           fds[2]          [out]       PIPE end ID
  [Return]
    Sub process ID
============================================================================= */
int execFilter(char *arg[], int fds[2])
{
  int child_pid = -1;
  
  DEBUG_PRINT_D("[rastertocanonijlfp] <Parent Process> Enter execFilter()\n");

  if (pipe(fds) >= 0)
  {
    child_pid = fork();

    if (child_pid == 0)
    {
      char shell_buf[BUF_SIZE] = {0};
      snprintf(shell_buf, BUF_SIZE, "%s/%s", PROG_PATH, CANONIJLFP_BIN);

      close(0);
      dup2(fds[0], 0);
      close(fds[0]);
      close(fds[1]);

      DEBUG_PRINT2_D("[rastertocanonijlfp] <Sub Process> Execute command: %s\n", shell_buf);

      execv(shell_buf, arg);

      DEBUG_PRINT_E("[rastertocanonijlfp] <Sub Process> execl() error\n");
    }
    else if (child_pid > 0)
    {
      DEBUG_PRINT2_D("[rastertocanonijlfp] <Parent Process> The sub process %d is forked\n", child_pid);
      close(fds[0]);
    }
  }

  DEBUG_PRINT_D("[rastertocanonijlfp] <Parent Process> Leave execFilter()\n");
  return child_pid;
}

/* =============================================================================
  [Function Name]
    sendData
  [Outline]
    Send raster data to sub process
  [Parameters]
    int           fdin       [in]         data receives from
    int           fdout      [in]         data sends to
  [Return]
    0: Succeed
    1: Failed
============================================================================= */
int sendData(int fdin, int fdout)
{
  char *p_data_buf;
  int read_bytes, write_bytes;
  int ret = 1;

  DEBUG_PRINT_D("[rastertocanonijlfp] <Parent Process> Enter sendData()\n");

  p_data_buf = malloc(DATA_BUF_SIZE * sizeof(char));

  while (!g_cancel_job)
  {
    read_bytes = read(fdin, p_data_buf, DATA_BUF_SIZE);
    if (read_bytes > 0)
    {
      char *p_data = p_data_buf;
      do
      {
        write_bytes = write(fdout, p_data, read_bytes);
        if (write_bytes < 0)
        {
          DEBUG_PRINT_E("[rastertocanonijlfp] <Parent Process> sendData() write error\n");
          goto EXIT;
        }
        read_bytes -= write_bytes;
        p_data += write_bytes;
      } while (!g_cancel_job && read_bytes > 0);
    }
    else if (read_bytes < 0)
    {
      DEBUG_PRINT_E("[rastertocanonijlfp] <Parent Process> sendData() read error\n");
      goto EXIT;
    }
    else
    {
      break;
    }
  }

  ret = 0;
  DEBUG_PRINT_D("[rastertocanonijlfp] <Parent Process> sendData(): Write data successfully\n");

EXIT:
  close(fdout);

  if(p_data_buf != NULL){
    free(p_data_buf);
  }

  DEBUG_PRINT_D("[rastertocanonijlfp] <Parent Process> Leave sendData()\n");
  return ret;
}

/* =============================================================================
  [Function Name]
    sigtermHandler
  [Outline]
    Process SIGTERM action
  [Parameters]
    int           sigcode       [in]         signal code
  [Return]
    Sub process ID
============================================================================= */
void sigtermHandler(int sigcode)
{
  g_cancel_job = 1;
  
  if (g_filter_pid != -1)
  {
    kill(-g_filter_pid, SIGTERM);
  }

  DEBUG_PRINT_D("[rastertocanonijlfp] sigtermHandler(): The process is terminated\n");
}
