/*
 *  CUPS add-on module for Canon Inkjet Printer.
 *  Copyright CANON INC. 2022
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
*/

#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/wait.h>
#include "com_def.h"
#include "com_log.h"
#include "common.h"

// Functions definiation
int g_cancel_job = 0;
int g_filter_pid = -1;

// Functions
/* =============================================================================
  [Function Name]
    main
  [Outline]
    Main function
  [Parameters]
    int           argc        [in]        arguments count
    char          *argv[]     [in]        arguments
  [Return]
    PROC_SUCCEEDED: Succeed
    PROC_FAILED: Failed
============================================================================= */
int main(int argc, char *argv[])
{
  struct sigaction sigact;
  cups_option_t *cups_option = NULL;
  int num_opt = 0;
  char **cmd_buf = NULL;
  int fds[2];
  int fdin = 0;
  int wret, wstatus;
  int ret = PROC_FAILED;

  DEBUG_PRINT_PROGRAM_START();

  // Ignore SIGPIPE
  memset(&sigact, 0, sizeof(sigact));
  sigact.sa_handler = SIG_IGN;
  if (sigaction(SIGPIPE, &sigact, NULL))
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] Failed to register SIGPIPE process method\n");
    goto EXIT;
  }

  // Set SIGTERM
  memset(&sigact, 0, sizeof(sigact));
  sigact.sa_handler = sigtermHandler;
  if (sigaction(SIGTERM, &sigact, NULL))
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] Failed to register SIGTERM process method\n");
    goto EXIT;
  }

  // Get print settings
  if (getPrintSettings(argv[5], &cups_option, &num_opt))
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] Failed to get print settings\n");
    goto EXIT;
  }

  if (getAdditionalPrintSettings(argv[2], argv[3], &cups_option, &num_opt))
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] Failed to get additional print settings\n");
    goto EXIT;
  }

  //cmd_buf[0] = canonijlfp
  //cmd_buf[1] = cups_option[0].name
  //cmd_buf[2] = cups_option[0].value
  //...
  //cmd_buf[end] = 0
  if((cmd_buf = calloc(num_opt * 2 + 2, sizeof(char *))) == NULL )
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] Not enough memory\n");
    goto EXIT;
  }

  // Make params for sub process
  if (makeExecCommand(cmd_buf, cups_option, num_opt))
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] Failed to make exec command\n");
    goto EXIT;
  }

  // Create PIPE and fork sub process
  g_filter_pid = execFilter(cmd_buf, fds);
  if (g_filter_pid == -1)
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] Failed to fork sub process\n");
    goto EXIT;
  }

  // Read raster data from stdin and send data to PIPE
  if (sendData(fdin, fds[1]))
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] Failed to send data to sub process\n");
    goto EXIT;
  }

  // Wait sub process to finish
  wret = waitpid( g_filter_pid, &wstatus, 0 );
  if (wret == -1 || !WIFEXITED(wstatus) || WEXITSTATUS(wstatus))
  {
    DEBUG_PRINT_E("[rastertocanonijlfp] Failed to execute sub process\n");
    goto EXIT;
  }

  ret = PROC_SUCCEEDED;
  DEBUG_PRINT_E("[rastertocanonijlfp] Process successfully\n");

EXIT:
  cupsFreeOptions(num_opt, cups_option);

  if (cmd_buf != NULL)
  {
    free(cmd_buf);
  }

  DEBUG_PRINT_PROGRAM_END();
  return ret;
}